from keras_bcr import BatchCorrRegularizer
import tensorflow as tf


# The BCR layer is added before the addition of the skip-connection
def build_resnet_block(inputs, units=64, activation="gelu",
                       dropout=0.4, bcr_rate=0.1):
    h = tf.keras.layers.Dense(units=units)(inputs)
    h = h = tf.keras.layers.Activation(activation=activation)(h)
    h = tf.keras.layers.Dropout(rate=dropout)(h)
    h = BatchCorrRegularizer(bcr_rate=bcr_rate)([h, inputs])  # << HERE
    outputs = tf.keras.layers.Add()([h, inputs])
    return outputs


# An model with 3 ResNet blocks
def build_model(input_dims):
    inputs = tf.keras.Input(shape=input_dims)
    h = build_resnet_block(inputs, units=input_dims[-1])
    h = build_resnet_block(h, units=input_dims[-1])
    outputs = build_resnet_block(h, units=input_dims[-1])
    model = tf.keras.Model(inputs=inputs, outputs=outputs)
    return model


def test1():
    INPUT_DIM = 64
    model = build_model(input_dims=[INPUT_DIM])
    model.compile(
        optimizer=tf.keras.optimizers.Adam(),
        loss="mean_squared_error")

    BATCH_SZ = 128
    X_train = tf.random.normal([BATCH_SZ, INPUT_DIM])
    y_train = tf.random.normal([BATCH_SZ])

    history = model.fit(X_train, y_train, verbose=1, epochs=2)
    assert "batch_corr_regularizer" in history.history.keys()


def test2():
    EMBED_DIM = 64
    SEQLEN = 32
    model = build_model(input_dims=[SEQLEN, EMBED_DIM])
    model.compile(
        optimizer=tf.keras.optimizers.Adam(),
        loss="mean_squared_error")

    BATCH_SZ = 128
    X_train = tf.random.normal([BATCH_SZ, SEQLEN, EMBED_DIM])
    y_train = tf.random.normal([BATCH_SZ, SEQLEN, EMBED_DIM])

    history = model.fit(X_train, y_train, verbose=1, epochs=2)
    assert "batch_corr_regularizer" in history.history.keys()



import tensorflow as tf


def batch_corr(a, b, tol: float = 1e-8):
    """ Average of absolute correlations coefficients
         for each neuron across a batch.

    Parameters:
    -----------
    a : tf.tensor
        The output of the ResNet layer before the skip connection

    b : tf.tensor
        The input for the ResNet Layer

    Example:
    --------
    import numpy as np
    BATCH_SIZE = 100
    NUM_NEURONS = 1024
    a = np.random.random((BATCH_SIZE, NUM_NEURONS))
    b = np.random.random((BATCH_SIZE, NUM_NEURONS))
    bcr = batch_corr(a, b)
    """
    # compute pearson' rhos
    mu_a = tf.reduce_mean(a, axis=0)
    mu_b = tf.reduce_mean(b, axis=0)
    centered_a = a - mu_a
    centered_b = b - mu_b
    nomin = tf.reduce_sum(tf.multiply(centered_a, centered_b), axis=0)
    denom1 = tf.math.sqrt(tf.reduce_sum(tf.math.pow(centered_a, 2), axis=0))
    denom2 = tf.math.sqrt(tf.reduce_sum(tf.math.pow(centered_b, 2), axis=0))
    rhos = nomin / (denom1 * denom2 + tol)
    # take the mean of absolute rhos
    return tf.math.reduce_mean(tf.math.abs(rhos))


class BatchCorrRegularizer(tf.keras.layers.Layer):
    """ Batch Correlation Regularizer

    Parameters:
    -----------
    bcr_rate : float (Default: 1e-6)
        The factor how much BCR regularization contributes to the loss function

    Example:
    --------
    The BatchCorrRegularizer must be integrated within a ResNet block. It
      cannot be attached between ResNet layers. Here is an example how to
      design a ResNet block using the Keras Functional API

    def ...(self, inputs, ...):
        h = tf.keras.Dense(...)(h)
        h = h = tf.keras.layers.Activation(...)(h)
        h = tf.keras.layers.Dropout(...)(h)
        h = BatchCorrRegularizer(bcr_rate)([h, inputs])
        outputs = tf.keras.layers.Add(...)([h, inputs])
        return outputs
    """
    def __init__(self, bcr_rate: float = 1e-6):
        super(BatchCorrRegularizer, self).__init__()
        self.bcr_rate = bcr_rate

    def call(self, inputs):
        lout, linp = inputs
        loss = self.bcr_rate * batch_corr(
            tf.keras.layers.Flatten()(lout),
            tf.keras.layers.Flatten()(linp)
        )
        self.add_loss(loss)
        self.add_metric(loss, name="batch_corr_regularizer")
        return lout



import setuptools
import os


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as fp:
        s = fp.read()
    return s


def get_version(path):
    with open(path, "r") as fp:
        lines = fp.read()
    for line in lines.split("\n"):
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name='keras-bcr',
    version=get_version("keras_bcr/__init__.py"),
    description='Batch Correlation Regularizer for TF2/Keras',
    long_description=read('README.rst'),
    url='http://github.com/satzbeleg/keras-bcr',
    author='Ulf Hamster',
    author_email='554c46@gmail.com',
    license='Apache License 2.0',
    packages=['keras_bcr'],
    install_requires=[
        "tensorflow>=2.4.0,<3"
    ],
    python_requires='>=3.7',
    zip_safe=True
)
